package at.ac.uibk.gitsearch.service.util;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.FileSystem;
import java.nio.file.InvalidPathException;
import java.nio.file.LinkOption;
import java.nio.file.Path;
import java.nio.file.ProviderMismatchException;
import java.nio.file.ReadOnlyFileSystemException;
import java.nio.file.WatchEvent;
import java.nio.file.WatchKey;
import java.nio.file.WatchService;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Objects;

/**
 * Base class for Path implementation for classic pathes with "/" and "." and
 * "..".
 *
 * @implNote This is mainly adapted from JrtPath
 */

@SuppressWarnings(
    {
        "PMD.AvoidUsingVolatile",
        "PMD.CyclomaticComplexity",
        "PMD.ControlStatementBraces",
        "PMD.GodClass",
        "PMD.TooManyMethods",
        "PMD.NPathComplexity",
        "PMD.CognitiveComplexity",
    }
)
public class SlashPath implements Path {

    private final String path;
    private volatile int[] offsets;
    private volatile String resolved = null;

    @SuppressWarnings("PMD.AvoidFieldNameMatchingMethodName")
    private boolean isDirectory = false;

    SlashPath(String path) {
        this.path = normalize(path);
        checkDirectory(path);
    }

    private void checkDirectory(String path) {
        if (path != null) {
            isDirectory = path.endsWith("/") || path.endsWith("/.") || path.endsWith("/..") || ".".equals(path) || "..".equals(path);
        }
    }

    SlashPath(String path, boolean normalized) {
        this.path = normalized ? path : normalize(path);
        checkDirectory(path);
    }

    final String getName() {
        return path;
    }

    @Override
    public final SlashPath getRoot() {
        return new SlashPath("/");
    }

    @Override
    public final SlashPath getFileName() {
        if (path.isEmpty()) return this;
        if (path.length() == 1 && path.charAt(0) == '/') return null;
        int off = path.lastIndexOf('/');
        if (off == -1) return this;
        return new SlashPath(path.substring(off + 1), true);
    }

    @Override
    public final SlashPath getParent() {
        initOffsets();
        int count = offsets.length;
        if (count == 0) { // no elements so no parent
            return getRoot();
        }
        int off = offsets[count - 1] - 1;
        if (off <= 0) { // parent is root only (may be null)
            return getRoot();
        }
        return new SlashPath(path.substring(0, off));
    }

    @Override
    public final int getNameCount() {
        initOffsets();
        return offsets.length;
    }

    @Override
    public final SlashPath getName(int index) {
        initOffsets();
        if (index < 0 || index >= offsets.length) {
            throw new IllegalArgumentException("index: " + index + ", offsets length: " + offsets.length);
        }
        int begin = offsets[index];
        int end;
        if (index == (offsets.length - 1)) {
            end = path.length();
        } else {
            end = offsets[index + 1];
        }
        return new SlashPath(path.substring(begin, end));
    }

    @Override
    public final SlashPath subpath(int beginIndex, int endIndex) {
        initOffsets();
        if (beginIndex < 0 || endIndex > offsets.length || beginIndex >= endIndex) {
            throw new IllegalArgumentException(
                "beginIndex: " + beginIndex + ", endIndex: " + endIndex + ", offsets length: " + offsets.length
            );
        }
        // starting/ending offsets
        int begin = offsets[beginIndex];
        int end;
        if (endIndex == offsets.length) {
            end = path.length();
        } else {
            end = offsets[endIndex];
        }
        return new SlashPath(path.substring(begin, end));
    }

    @Override
    public final SlashPath toRealPath(LinkOption... options) throws IOException {
        return toAbsolutePath();
    }

    @Override
    public final SlashPath toAbsolutePath() {
        if (isAbsolute()) return this;
        return new SlashPath("/" + path, true);
    }

    @Override
    public final URI toUri() {
        try {
            return new URI("jrt", toAbsolutePath().path, null);
        } catch (URISyntaxException ex) {
            throw new AssertionError(ex);
        }
    }

    private boolean equalsNameAt(SlashPath other, int index) {
        int mbegin = offsets[index];
        int mlen;
        if (index == (offsets.length - 1)) {
            mlen = path.length() - mbegin;
        } else {
            mlen = offsets[index + 1] - mbegin - 1;
        }
        int obegin = other.offsets[index];
        int olen;
        if (index == (other.offsets.length - 1)) {
            olen = other.path.length() - obegin;
        } else {
            olen = other.offsets[index + 1] - obegin - 1;
        }
        if (mlen != olen) {
            return false;
        }
        int n = 0;
        while (n < mlen) {
            if (path.charAt(mbegin + n) != other.path.charAt(obegin + n)) {
                return false;
            }
            n++;
        }
        return true;
    }

    @Override
    public final SlashPath relativize(Path other) {
        final SlashPath o = checkPath(other);
        if (o.equals(this)) {
            return new SlashPath("", true);
        }
        if (path.isEmpty()) {
            return o;
        }
        if (isAbsolute() != o.isAbsolute()) {
            throw new IllegalArgumentException("Incorrect filesystem or path: " + other);
        }
        final String tp = this.path;
        final String op = o.path;
        if (op.startsWith(tp)) { // fast path
            int off = tp.length();
            if (op.charAt(off - 1) == '/') return new SlashPath(op.substring(off), true);
            if (op.charAt(off) == '/') return new SlashPath(op.substring(off + 1), true);
        }
        int mc = this.getNameCount();
        int oc = o.getNameCount();
        int n = Math.min(mc, oc);
        int i = 0;
        while (i < n) {
            if (!equalsNameAt(o, i)) {
                break;
            }
            i++;
        }
        int dotdots = mc - i;
        int len = dotdots * 3 - 1;
        if (i < oc) {
            len += (o.path.length() - o.offsets[i] + 1);
        }
        StringBuilder sb = new StringBuilder(len);
        while (dotdots > 0) {
            sb.append("..");
            if (sb.length() < len) { // no tailing slash at the end
                sb.append('/');
            }
            dotdots--;
        }
        if (i < oc) {
            sb.append(o.path, o.offsets[i], o.path.length());
        }
        return new SlashPath(sb.toString(), true);
    }

    @Override
    public FileSystem getFileSystem() {
        return null;
    }

    @Override
    public final boolean isAbsolute() {
        return !path.isEmpty() && path.charAt(0) == '/';
    }

    @Override
    public final SlashPath resolve(Path other) {
        final SlashPath o = checkPath(other);
        if (this.path.isEmpty() || o.isAbsolute()) {
            return o;
        }
        if (o.path.isEmpty()) {
            return this;
        }
        StringBuilder sb = new StringBuilder(path.length() + o.path.length() + 1);
        sb.append(path);
        if (path.charAt(path.length() - 1) != '/') sb.append('/');
        sb.append(o.path);
        final SlashPath resolvedPath = new SlashPath(sb.toString(), true);
        resolvedPath.isDirectory = o.isDirectory;
        return resolvedPath;
    }

    @Override
    public final Path resolveSibling(Path other) {
        Objects.requireNonNull(other, "other");
        Path parent = getParent();
        return (parent == null) ? other : parent.resolve(other);
    }

    @Override
    public final boolean startsWith(Path other) {
        if (!(Objects.requireNonNull(other) instanceof SlashPath)) return false;
        final SlashPath o = (SlashPath) other;
        final String tp = this.path;
        final String op = o.path;
        if (isAbsolute() != o.isAbsolute() || !tp.startsWith(op)) {
            return false;
        }
        int off = op.length();
        if (off == 0) {
            return tp.isEmpty();
        }
        // check match is on name boundary
        return tp.length() == off || tp.charAt(off) == '/' || off == 0 || op.charAt(off - 1) == '/';
    }

    @Override
    @SuppressWarnings({ "PMD.NPathComplexity", "PMD.UseEqualsToCompareStrings", "PMD.CompareObjectsWithEquals" })
    public final boolean endsWith(Path other) {
        if (!(Objects.requireNonNull(other) instanceof SlashPath)) return false;
        final SlashPath o = (SlashPath) other;
        final SlashPath t = this;
        int olast = o.path.length() - 1;
        if (olast > 0 && o.path.charAt(olast) == '/') {
            olast--;
        }
        int last = t.path.length() - 1;
        if (last > 0 && t.path.charAt(last) == '/') {
            last--;
        }
        if (olast == -1) { // o.path.length == 0
            return last == -1;
        }
        if ((o.isAbsolute() && (!t.isAbsolute() || olast != last)) || last < olast) {
            return false;
        }
        for (; olast >= 0; olast--, last--) {
            if (o.path.charAt(olast) != t.path.charAt(last)) {
                return false;
            }
        }
        return o.path.charAt(olast + 1) == '/' || last == -1 || t.path.charAt(last) == '/';
    }

    @Override
    public final SlashPath resolve(String other) {
        return resolve(getPath(other));
    }

    @Override
    public final Path resolveSibling(String other) {
        return resolveSibling(getPath(other));
    }

    @Override
    public final boolean startsWith(String other) {
        return startsWith(getPath(other));
    }

    @Override
    public final boolean endsWith(String other) {
        return endsWith(getPath(other));
    }

    @Override
    @SuppressWarnings({ "PMD.UseEqualsToCompareStrings", "PMD.CompareObjectsWithEquals" })
    public final SlashPath normalize() {
        String res = getResolved();
        if (res == path) { // no change
            return this;
        }
        return new SlashPath(res, true);
    }

    private SlashPath checkPath(Path path) {
        Objects.requireNonNull(path);
        if (!(path instanceof SlashPath)) throw new ProviderMismatchException("path class: " + path.getClass());
        return (SlashPath) path;
    }

    // create offset list if not already created
    private void initOffsets() {
        if (this.offsets == null) {
            int len = path.length();
            // count names
            int count = 0;
            int off = 0;
            while (off < len) {
                char c = path.charAt(off++);
                if (c != '/') {
                    count++;
                    off = path.indexOf('/', off);
                    if (off == -1) break;
                }
            }
            // populate offsets
            int[] offsets = new int[count];
            count = 0;
            off = 0;
            while (off < len) {
                char c = path.charAt(off);
                if (c == '/') {
                    off++;
                } else {
                    offsets[count++] = off++;
                    off = path.indexOf('/', off);
                    if (off == -1) break;
                }
            }
            this.offsets = offsets;
        }
    }

    final String getResolvedPath() {
        String r = resolved;
        if (r == null) {
            if (isAbsolute()) {
                r = getResolved();
            } else {
                r = toAbsolutePath().getResolvedPath();
            }
            resolved = r;
        }
        return r;
    }

    // removes redundant slashs, replace "\" to separator "/"
    // and check for invalid characters
    private static String normalize(String path) {
        int len = path.length();
        if (len == 0) {
            return path;
        }
        char prevC = 0;
        for (int i = 0; i < len; i++) {
            char c = path.charAt(i);
            if (c == '\\' || c == '\u0000') {
                return normalize(path, i);
            }
            if (c == '/' && prevC == '/') {
                return normalize(path, i - 1);
            }
            prevC = c;
        }
        if (prevC == '/' && len > 1) {
            return path.substring(0, len - 1);
        }
        return path;
    }

    private static String normalize(String path, int off) {
        int len = path.length();
        StringBuilder to = new StringBuilder(len);
        to.append(path, 0, off);
        char prevC = 0;
        while (off < len) {
            @SuppressWarnings("PMD.AvoidReassigningParameters")
            char c = path.charAt(off++);
            if (c == '\\') {
                c = '/';
            }
            if (c == '/' && prevC == '/') {
                continue;
            }
            if (c == '\u0000') {
                throw new InvalidPathException(path, "Path: NUL character not allowed");
            }
            to.append(c);
            prevC = c;
        }
        len = to.length();
        if (len > 1 && to.charAt(len - 1) == '/') {
            to.deleteCharAt(len - 1);
        }
        return to.toString();
    }

    // Remove DotSlash(./) and resolve DotDot (..) components
    private String getResolved() {
        int length = path.length();
        if (length == 0 || (path.indexOf("./") == -1 && path.charAt(length - 1) != '.')) {
            return path;
        } else {
            return resolvePath();
        }
    }

    @SuppressWarnings("PMD.AssignmentInOperand")
    private String resolvePath() {
        int length = path.length();
        char[] to = new char[length];
        int nc = getNameCount();
        int[] lastM = new int[nc];
        int lastMOff = -1;
        int m = 0;
        for (int i = 0; i < nc; i++) {
            int n = offsets[i];
            int len = (i == offsets.length - 1) ? length - n : offsets[i + 1] - n - 1;
            if (len == 1 && path.charAt(n) == '.') {
                if (m == 0 && path.charAt(0) == '/') to[m++] = '/'; // absolute path
                continue;
            }
            if (len == 2 && path.charAt(n) == '.' && path.charAt(n + 1) == '.') {
                if (lastMOff >= 0) {
                    m = lastM[lastMOff--]; // retreat
                    continue;
                }
                if (path.charAt(0) == '/') { // "/../xyz" skip
                    if (m == 0) to[m++] = '/';
                } else { // "../xyz" -> "../xyz"
                    if (m != 0 && to[m - 1] != '/') to[m++] = '/';
                    while (len-- > 0) {
                        to[m++] = path.charAt(n++);
                    }
                }
                continue;
            }
            if (
                m == 0 &&
                path.charAt(0) == '/' || // absolute path
                m != 0 &&
                to[m - 1] != '/'
            ) { // not the first name
                to[m++] = '/';
            }
            lastM[++lastMOff] = m;
            while (len-- > 0) to[m++] = path.charAt(n++);
        }
        if (m > 1 && to[m - 1] == '/') m--;
        return (m == to.length) ? new String(to) : new String(to, 0, m);
    }

    @Override
    public final String toString() {
        return path;
    }

    @Override
    public final int hashCode() {
        return path.hashCode();
    }

    @Override
    public final boolean equals(Object obj) {
        return obj instanceof SlashPath && this.path.equals(((SlashPath) obj).path);
    }

    @Override
    public final int compareTo(Path other) {
        final SlashPath o = checkPath(other);
        return path.compareTo(o.path);
    }

    @Override
    public final WatchKey register(WatchService watcher, WatchEvent.Kind<?>[] events, WatchEvent.Modifier... modifiers) {
        Objects.requireNonNull(watcher, "watcher");
        Objects.requireNonNull(events, "events");
        Objects.requireNonNull(modifiers, "modifiers");
        throw new UnsupportedOperationException();
    }

    @Override
    public final WatchKey register(WatchService watcher, WatchEvent.Kind<?>... events) {
        return register(watcher, events, new WatchEvent.Modifier[0]);
    }

    @Override
    public final File toFile() {
        throw new UnsupportedOperationException();
    }

    @Override
    public final Iterator<Path> iterator() {
        return new Iterator<Path>() {
            private int i = 0;

            @Override
            public boolean hasNext() {
                return (i < getNameCount());
            }

            @Override
            public Path next() {
                if (i < getNameCount()) {
                    Path result = getName(i);
                    i++;
                    return result;
                } else {
                    throw new NoSuchElementException();
                }
            }

            @Override
            public void remove() {
                throw new ReadOnlyFileSystemException();
            }
        };
    }

    public SlashPath getPath(String first, String... more) {
        if (more.length == 0) {
            return new SlashPath(first);
        }
        StringBuilder sb = new StringBuilder();
        sb.append(first);
        for (String path : more) {
            if (!path.isEmpty()) {
                if (sb.length() > 0) {
                    sb.append('/');
                }
                sb.append(path);
            }
        }
        return new SlashPath(sb.toString());
    }

    /**
     * Converts a path string, or a sequence of strings that when joined form
     * a path string, to a {@code Path}.
     *
     * @implSpec
     *           This method simply invokes {@link Path#of(String,String...)
     *           Path.of(String, String...)} with the given parameters.
     *
     * @param first
     *              the path string or initial part of the path string
     * @param more
     *              additional strings to be joined to form the path string
     *
     * @return the resulting {@code Path}
     *
     * @throws InvalidPathException
     *                              if the path string cannot be converted to a
     *                              {@code Path}
     *
     * @see FileSystem#getPath
     * @see Path#of(String,String...)
     */
    public static SlashPath get(String first, String... more) {
        if (more.length == 0) {
            return new SlashPath(first);
        }
        StringBuilder sb = new StringBuilder();
        sb.append(first);
        for (String path : more) {
            if (!path.isEmpty()) {
                if (sb.length() > 0) {
                    sb.append('/');
                }
                sb.append(path);
            }
        }
        return new SlashPath(sb.toString());
    }

    public boolean isDirectory() {
        return isDirectory;
    }
}
