import { ComponentFixture, TestBed } from '@angular/core/testing';
import { HttpResponse } from '@angular/common/http';
import { HttpClientTestingModule } from '@angular/common/http/testing';
import { FormBuilder } from '@angular/forms';
import { ActivatedRoute } from '@angular/router';
import { RouterTestingModule } from '@angular/router/testing';
import { of, Subject, from } from 'rxjs';

import { WatchListEntryService } from '../service/watch-list-entry.service';
import { IWatchListEntry, WatchListEntry } from '../watch-list-entry.model';
import { IUserWatchList } from 'app/entities/user-watch-list/user-watch-list.model';
import { UserWatchListService } from 'app/entities/user-watch-list/service/user-watch-list.service';

import { WatchListEntryUpdateComponent } from './watch-list-entry-update.component';

describe('WatchListEntry Management Update Component', () => {
  let comp: WatchListEntryUpdateComponent;
  let fixture: ComponentFixture<WatchListEntryUpdateComponent>;
  let activatedRoute: ActivatedRoute;
  let watchListEntryService: WatchListEntryService;
  let userWatchListService: UserWatchListService;

  beforeEach(() => {
    TestBed.configureTestingModule({
      imports: [HttpClientTestingModule, RouterTestingModule.withRoutes([])],
      declarations: [WatchListEntryUpdateComponent],
      providers: [
        FormBuilder,
        {
          provide: ActivatedRoute,
          useValue: {
            params: from([{}]),
          },
        },
      ],
    })
      .overrideTemplate(WatchListEntryUpdateComponent, '')
      .compileComponents();

    fixture = TestBed.createComponent(WatchListEntryUpdateComponent);
    activatedRoute = TestBed.inject(ActivatedRoute);
    watchListEntryService = TestBed.inject(WatchListEntryService);
    userWatchListService = TestBed.inject(UserWatchListService);

    comp = fixture.componentInstance;
  });

  describe('ngOnInit', () => {
    it('Should call UserWatchList query and add missing value', () => {
      const watchListEntry: IWatchListEntry = { id: 456 };
      const watchlist: IUserWatchList = { id: 81709 };
      watchListEntry.watchlist = watchlist;

      const userWatchListCollection: IUserWatchList[] = [{ id: 15755 }];
      jest.spyOn(userWatchListService, 'query').mockReturnValue(of(new HttpResponse({ body: userWatchListCollection })));
      const additionalUserWatchLists = [watchlist];
      const expectedCollection: IUserWatchList[] = [...additionalUserWatchLists, ...userWatchListCollection];
      jest.spyOn(userWatchListService, 'addUserWatchListToCollectionIfMissing').mockReturnValue(expectedCollection);

      activatedRoute.data = of({ watchListEntry });
      comp.ngOnInit();

      expect(userWatchListService.query).toHaveBeenCalled();
      expect(userWatchListService.addUserWatchListToCollectionIfMissing).toHaveBeenCalledWith(
        userWatchListCollection,
        ...additionalUserWatchLists
      );
      expect(comp.userWatchListsSharedCollection).toEqual(expectedCollection);
    });

    it('Should update editForm', () => {
      const watchListEntry: IWatchListEntry = { id: 456 };
      const watchlist: IUserWatchList = { id: 13918 };
      watchListEntry.watchlist = watchlist;

      activatedRoute.data = of({ watchListEntry });
      comp.ngOnInit();

      expect(comp.editForm.value).toEqual(expect.objectContaining(watchListEntry));
      expect(comp.userWatchListsSharedCollection).toContain(watchlist);
    });
  });

  describe('save', () => {
    it('Should call update service on save for existing entity', () => {
      // GIVEN
      const saveSubject = new Subject<HttpResponse<WatchListEntry>>();
      const watchListEntry = { id: 123 };
      jest.spyOn(watchListEntryService, 'update').mockReturnValue(saveSubject);
      jest.spyOn(comp, 'previousState');
      activatedRoute.data = of({ watchListEntry });
      comp.ngOnInit();

      // WHEN
      comp.save();
      expect(comp.isSaving).toEqual(true);
      saveSubject.next(new HttpResponse({ body: watchListEntry }));
      saveSubject.complete();

      // THEN
      expect(comp.previousState).toHaveBeenCalled();
      expect(watchListEntryService.update).toHaveBeenCalledWith(watchListEntry);
      expect(comp.isSaving).toEqual(false);
    });

    it('Should call create service on save for new entity', () => {
      // GIVEN
      const saveSubject = new Subject<HttpResponse<WatchListEntry>>();
      const watchListEntry = new WatchListEntry();
      jest.spyOn(watchListEntryService, 'create').mockReturnValue(saveSubject);
      jest.spyOn(comp, 'previousState');
      activatedRoute.data = of({ watchListEntry });
      comp.ngOnInit();

      // WHEN
      comp.save();
      expect(comp.isSaving).toEqual(true);
      saveSubject.next(new HttpResponse({ body: watchListEntry }));
      saveSubject.complete();

      // THEN
      expect(watchListEntryService.create).toHaveBeenCalledWith(watchListEntry);
      expect(comp.isSaving).toEqual(false);
      expect(comp.previousState).toHaveBeenCalled();
    });

    it('Should set isSaving to false on error', () => {
      // GIVEN
      const saveSubject = new Subject<HttpResponse<WatchListEntry>>();
      const watchListEntry = { id: 123 };
      jest.spyOn(watchListEntryService, 'update').mockReturnValue(saveSubject);
      jest.spyOn(comp, 'previousState');
      activatedRoute.data = of({ watchListEntry });
      comp.ngOnInit();

      // WHEN
      comp.save();
      expect(comp.isSaving).toEqual(true);
      saveSubject.error('This is an error!');

      // THEN
      expect(watchListEntryService.update).toHaveBeenCalledWith(watchListEntry);
      expect(comp.isSaving).toEqual(false);
      expect(comp.previousState).not.toHaveBeenCalled();
    });
  });

  describe('Tracking relationships identifiers', () => {
    describe('trackUserWatchListById', () => {
      it('Should return tracked UserWatchList primary key', () => {
        const entity = { id: 123 };
        const trackResult = comp.trackUserWatchListById(0, entity);
        expect(trackResult).toEqual(entity.id);
      });
    });
  });
});
